# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Function registry to map function names to implementations for page
    deserialization (in case the import is not consistent from serialization
    to deserialization)
"""


class FunctionRegistry(object):
    """ Provides function import translation for de-serialization.

    A glorified dictionary for now.
    """

    def __init__(self):
        self.registry = {}

    def register(self, function):
        """ Register a function for deserialization

        Parameters
        ----------
        function : Callable
            The function that implements the name, to allow a function to
            be found separately than the original module from serialization.
        """
        name = function.__name__
        if name in self.registry:
            raise ValueError(f"The function {name!r} is already registered")
        self.registry[name] = function

    def retrieve(self, name):
        """ Retrieve the callable for the given name.

        Parameters
        ----------
        name : String
            The name associated with the callable

        Returns
        -------
        Callable
            The function that implements the name, or None
        """
        return self.registry.get(name)

    def list_names(self):
        """ List all known names.
        More useful for debugging than anything else.

        Returns
        -------
        List(String)
        """
        return sorted(self.registry)



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Classes to represent functions for widgets, craeted to hold data for
    serialization.
"""
from abc import ABC
from copy import deepcopy


class IFunction(ABC):
    """ Generic representation of a callable function, providing a layer
    of storage and indirection for serialization.
    """

    def __init__(self, function):
        """ Init the generic function with a callable.

        Parameters
        ----------
        function : Callable
        """
        self.function = function

    def execute(self, **kwargs):
        """ Execute the generic function with generic parameters.

        Parameters
        ----------
        **kwargs : Dict

        Returns
        -------
        Any
        """
        return self.function(**kwargs)


class ButtonFunction(IFunction):
    """ A representation of a function that is useful for a button.
    """

    def __init__(self, function, data=None):
        """ Initialize the button function, with optional data that will
        be serialized with the function.

        Parameters
        ----------
        function : Callable
            The function to be called when the button changes state. Will
            be called with kwargs as a dictionary, containing `page`, `button`
            and `state` plus whatever is stored in `base_kwargs`
        data : dict
            Optional dictionary entries to be passed to the button function.
        """
        self.data = data or {}
        super(ButtonFunction, self).__init__(function)

    def update(self, page, button, state):
        """ Update the button state (e.g. for push, release, etc.).

        If the update returns a Page, that page will be the new page
        rendered.

        Parameters
        ----------
        page : Page
            The page the button is on
        button : IButtonWidget
            The button object calling update
        state : Boolean
            The state of the button, True being pushed

        Returns
        -------
        Optional['Page']
        """
        kwargs = deepcopy(self.data)
        kwargs.update({'page': page, 'button': button, 'state': state})

        return self.execute(**kwargs)


class ButtonGroupFunction(ButtonFunction):
    """ A representation of a function that is useful for a button.
    """

    def __init__(self, function, data=None):
        super(ButtonGroupFunction, self).__init__(function, data)


class IControlFunction(IFunction):
    """ A representation of a function that is useful for an input control.
    """

    def __init__(self, function, data=None):
        """ Initialize the control function, with optional data that will
        be serialized with the function.

        Parameters
        ----------
        function : Callable
            The function to be called when the control changes value. Will
            be called with kwargs as a dictionary, containing `page`,
            `control` and `value` plus whatever is stored in `base_kwargs`
        data : dict
            Optional dictionary entries to be passed to the control function.
        """
        self.data = data or {}
        super(IControlFunction, self).__init__(function)

    def update(self, page, control, value):
        """ Update the control value

        Parameters
        ----------
        page : Page
            The page the control is on
        control : ControlWidget
            The button object calling update
        value : Number
            The value of the control
        """
        kwargs = deepcopy(self.data)
        kwargs.update({'page': page, 'control': control, 'value': value})

        return self.execute(**kwargs)


class DialFunction(IControlFunction):
    def __init__(self, function, data=None):
        super(DialFunction, self).__init__(function, data)


class SliderFunction(IControlFunction):
    def __init__(self, function, data=None):
        super(SliderFunction, self).__init__(function, data)


class IndicatorFunction(IFunction):
    """ A representation of a function that is useful for an indicator (an
    active label).
    """

    def __init__(self, function, data=None):
        """ Initialize the indicator function, with optional data that will
        be serialized with the function.

        Parameters
        ----------
        function : Callable
            The function to be called to retrieve the indicator status. Will
            be called with kwargs as a dictionary, containing `page`,
            `indicator`, 'label', and 'active', plus whatever is stored in
            `base_kwargs`
        data : dict
            Optional dictionary entries to be passed to the control function.
        """
        self.data = data or {}
        super(IndicatorFunction, self).__init__(function)

    def status(self, page, indicator, label, active):
        """ Retrieve the indicator status.

        Parameters
        ----------
        page : Page
            The page the indicator is on
        indicator : IndicatorWidget
            The indicator object calling status
        label : String
            The base label of the indicator
        active : Boolean
            True if the indicator is considered "active"

        Returns
        -------
        (Bool, String)
            The bool is an 'active' flag, which sets the active color
            palette if True.  The string sets the indicator label.  If it is
            None, then the indicator defaults to self.label.
        """
        kwargs = deepcopy(self.data)
        kwargs.update({'page': page,
                       'indicator': indicator,
                       'label': label,
                       'active': active})

        return self.execute(**kwargs)


class LightFunction(IFunction):
    """ A representation of a function that is useful for the display light
    """

    def __init__(self, function, data=None):
        """ Initialize the light function, with optional data that will
        be serialized with the function.

        Parameters
        ----------
        function : Callable
            The function to be called to get the light colors. Will
            be called with kwargs as a dictionary, containing `page`, `light`
            plus whatever is stored in `base_kwargs`
        data : dict
            Optional dictionary entries to be passed to the control function.
        """
        self.data = data or {}
        super(LightFunction, self).__init__(function)

    def colors(self, page, light):
        """ Retrieve the light colors.

        Parameters
        ----------
        page : Page
            The page the light is on
        light : LightWidget
            The light object calling update

        Returns
        -------
        (Color, Color)
            The (ring, fill) colors
        """
        kwargs = deepcopy(self.data)
        kwargs.update({'page': page, 'light': light})

        return self.execute(**kwargs)


class GraphFunction(IFunction):
    """ A representation of a function that is useful for an input control.
    """

    def __init__(self, function, data=None):
        """ Initialize the button function, with optional data that will
        be serialized with the function.

        Parameters
        ----------
        function : Callable
            The function to be called to request values. Will
            be called with kwargs as a dictionary, containing `page`, `graph`
            and `num` plus whatever is stored in `base_kwargs`
        data : dict
            Optional dictionary entries to be passed to the control function.
        """
        self.data = data or {}
        super(GraphFunction, self).__init__(function)

    def samples(self, page, graph, num):
        """ Retrieve some samples

        Parameters
        ----------
        page : Page
            The page the graph is on.
        graph : GraphWidget
            The button object calling update
        num : Int
            The number of samples to retrieve

        Returns
        -------
        List[Dict]
            One dictionary per trace in the graph.
            The dictionary defines the color of the trace and a list of float
            with the sample data:
                {"color": <Color>, "data": [<float>]}

        """
        kwargs = deepcopy(self.data)
        kwargs.update({'page': page, 'graph': graph, 'num': num})

        return self.execute(**kwargs)



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Graphical User Interface objects

    A `Page` is a set of widgets that define the interaction and display
    surface of the GUI

    The `GUI` manages the hardware and runs the Pages
"""
import importlib
import json
import logging
from pathlib import Path
from typing import List, Optional

import pygame

from pi_touch_gui._utilities import backlog_error
from pi_touch_gui._widget_bases import IWidget, IWidgetInterface
from pi_touch_gui.colors import BLACK
from pi_touch_gui.widget_functions import IFunction

LOG = logging.getLogger(__name__)


class Page(object):
    LCARS = 'lcars'

    def __init__(self, name, widgets=None, background=None):
        """ A collection of widgets that make up an interaction page.

        Parameters
        ----------
        name : String
            Name of the page for internal and callback reference
        widgets : List[IWidget]
        background : str
            Background image resource filepath for this Page
        """
        # If no widgets added on init, they can be added with `add_widgets()`
        # later.
        self.name = name
        self.widgets = widgets or []

        if background == Page.LCARS:
            background = str(Path(Path(__file__).parent,
                                  "assets/lcars_screen.png"))
        self.background = background
        self._bg_image = None

        if widgets is None:
            self.widgets = []
        else:
            self.widgets = widgets
        self._widget_data = {}

        self._last_adjustment = pygame.time.get_ticks()
        self._adjustment_rate = 100

    @property
    def widget_data(self):
        # Global named data store for widgets to communicate
        return self._widget_data

    def set_background(self, filepath):
        self.background = filepath
        self._bg_image = pygame.image.load(filepath).convert()

    def add_widgets(self, widgets):
        """ Add widgets to the Page.

        Parameters
        ----------
        widgets : List[IWidget]
        """
        self.widgets += widgets

    def touch_handler(self, event, touch) -> Optional['Page']:
        """ Update all widgets with a touch event.

            The first widget to consume the touch wins, and downstream
            widgets won't be tested.

            If no widget consumes the event, we call `function(self) -> Page`

        Parameters
        ----------
        event : int
            Event code defined for the interface; e.g. _widget_bases.py imports
            TS_PRESS, TS_RELEASE, and TS_MOVE
        touch : Touch
            The Touch object for the event.

        Returns
        -------
        Optional[Page]
            If the widget function names a new page, it is returned here
        """
        # Give the widgets the page context
        touch.page = self

        for widget in self.widgets:
            consumed, new_page = widget.event(event, touch)
            if consumed:
                return new_page

        return None

    def render(self, surface: pygame.Surface):
        """Redraw all widgets to screen.

        Call `render()` for all widgets in the Page, once per frame.

        Parameters
        ----------
        surface : pygame.Surface
            Surface (screen) to draw to
        """
        # Lazy loading to avoid pygame initialization issues
        if self._bg_image is None and self.background is not None:
            self._bg_image = pygame.image.load(self.background).convert()

        if self._bg_image is not None:
            surface.blit(self._bg_image, (0, 0))
        else:
            surface.fill(BLACK)
        for widget in self.widgets:
            try:
                widget.render(self, surface)
            except Exception as e:
                backlog_error(e, f"Problem rendering widget {widget}")
                raise

    def iter_widgets(self):
        """ Iterate over the widgets in the page.

        Returns
        -------
        iter(IWidget)
        """
        return iter(self.widgets)

    def serialize(self, readable=False):
        """ Serialize this page and return it as a JSON string, suitable
        to be deserialized()

        Parameters
        ----------
        readable : Bool
            True if we return formatted json

        Returns
        -------
        String
            json representation of this Page
        """
        indent = 4 if readable else None
        functions = set()

        serial_page = json.dumps(self._do_serialize(self, functions),
                          indent=indent,
                          sort_keys=True)
        LOG.info(f"Page {self.name!r} uses:  {functions}")
        return serial_page

    def list_functions(self):
        functions = set()
        self._do_serialize(self, functions)
        return functions

    def _do_serialize(self, entity, functions=None):
        entity_type = type(entity)

        if entity is None:
            return None

        if entity_type in [str, int, float, complex, bool]:
            return entity

        elif entity_type in [list, tuple, set]:
            # JSON doesn't distinguish list-equivalent types
            return [self._do_serialize(v, functions)
                    for v in entity]

        elif entity_type is dict:
            return {k: self._do_serialize(v, functions)
                    for k, v in entity.items()}

        elif callable(entity):
            path = '.'.join([entity.__module__,
                             entity.__qualname__]).split('.')
            context = '.'.join(path[:-1])
            name = path[-1]

            if '<locals>' in context:
                raise ValueError(f'Can not serialize local function '
                                 f'{context}.{name}')
            if '<lambda>' in context:
                raise ValueError(f'Can not serialize lambda function '
                                 f'{context}.{name}')
            elif 'builtins' in context:
                raise ValueError(f'Can not serialize builtin function '
                                 f'{context}.{name}')

            if functions is not None:
                functions.add(name)
            return {'__Function': {'context': context,
                                   'name': name}}

        elif issubclass(entity_type, Page):
            return self._do_serialize_object(entity, functions)

        elif issubclass(entity_type, IWidgetInterface):
            return self._do_serialize_object(entity, functions)

        elif issubclass(entity_type, IFunction):
            return self._do_serialize_object(entity, functions)

        else:
            raise TypeError(f"Unsupported type {entity_type!r}")

    def _do_serialize_object(self, entity, functions):
        ent_type = type(entity)
        kwargs = {k: self._do_serialize(v, functions)
                  for k, v in vars(entity).items()
                  if not k.startswith('_')}
        return {'__Object': {'module': ent_type.__module__,
                             'class': ent_type.__name__,
                             'kwargs': kwargs}}

    @classmethod
    def deserialize(cls, serial, registry=None):
        """ Deserialize the data and return the Page it started as.

        Parameters
        ----------
        serial : String
            Serialized page
        registry : FunctionRegistry
            Registry of functions to replace normal imports

        Returns
        -------
        Page
        """
        data = json.loads(serial)
        return Page._do_deserialize(registry, data)
        pass

    @classmethod
    def _do_deserialize(cls, registry, entity):
        entity_type = type(entity)

        if entity is None:
            return None
        if entity_type in [str, int, float, complex, bool]:
            return entity
        if entity_type is list:
            return [Page._do_deserialize(registry, v) for v in entity]
        elif entity_type is dict:
            content = {}
            for key in entity.keys():
                if key == '__Function':
                    value = entity[key]
                    context = value['context']
                    func_name = value['name']
                    return Page._get_context(registry, context, func_name)
                elif key == '__Object':
                    value = entity[key]
                    module = value['module']
                    class_name = value['class']
                    kwargs = Page._do_deserialize(registry, value['kwargs'])
                    c = Page._get_context(registry, module, class_name)
                    return c(**kwargs)
                else:
                    content.update(
                        {key: Page._do_deserialize(registry, entity[key])})
            return content
        else:
            raise TypeError(f"Unsupported type {entity_type!r}")

    @classmethod
    def _get_context(cls, registry, module, item):
        if registry:
            function =  registry.retrieve(item)
            if function is not None:
                return function

        path = module.split('.')
        imp = path[0]
        try:
            m = importlib.import_module(imp)
        except Exception as e:
            raise ImportError(f"Unable to import {module!r}.{item}: {e}")
        for part in path[1:]:
            m = getattr(m, part)
        return getattr(m, item)



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Touchscreen for the official Raspberry PI 4's 7" touchscreen hardware.

    Based VERY CLOSELY on the Pimoroni hp4ts.py (see the Pimoroni folder)
"""
import errno
import glob
import io
import queue
import select
import struct
import threading
from collections import namedtuple
from pathlib import Path

import pygame

TOUCH_X = 0
TOUCH_Y = 1

TouchEvent = namedtuple('TouchEvent', ('timestamp', 'type', 'code', 'value'))

EV_SYN = 0
EV_ABS = 3

ABS_X = 0
ABS_Y = 1

ABS_MT_SLOT = 0x2f  # 47 MT slot being modified
ABS_MT_POSITION_X = 0x35  # 53 Center X of multi touch position
ABS_MT_POSITION_Y = 0x36  # 54 Center Y of multi touch position
ABS_MT_TRACKING_ID = 0x39  # 57 Unique ID of initiated contact

# Touch events
TS_PRESS = 1
TS_RELEASE = 0
TS_MOVE = 2


class Touch(object):
    def __init__(self, slot, x, y):
        self.slot = slot

        self._x = x
        self._y = y
        self.last_x = -1
        self.last_y = -1

        self._id = -1
        self.events = []
        self._on_move = None
        self._on_press = None
        self._on_release = None
        self._page = None

    @property
    def page(self):
        return self._page

    @page.setter
    def page(self, page):
        self._page = page

    @property
    def position(self):
        return self.x, self.y

    @property
    def last_position(self):
        return self.last_x, self.last_y

    @property
    def valid(self):
        return self.id > -1

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        if value != self._id:
            if value == -1 and TS_RELEASE not in self.events:
                self.events.append(TS_RELEASE)
            elif TS_PRESS not in self.events:
                self.events.append(TS_PRESS)

        self._id = value

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, value):
        if value != self._x and TS_MOVE not in self.events:
            self.events.append(TS_MOVE)
        self.last_x = self._x
        self._x = value

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, value):
        if value != self._y and TS_MOVE not in self.events:
            self.events.append(TS_MOVE)
        self.last_y = self._y
        self._y = value

    def handle_events(self):
        """Run outstanding press/release/move events"""
        for event in self.events:
            if event == TS_MOVE and callable(self._on_move):
                self._on_move(event, self)
            if event == TS_PRESS and callable(self._on_press):
                self._on_press(event, self)
            if event == TS_RELEASE and callable(self._on_release):
                self._on_release(event, self)

        self.events = []


class Touches(list):
    @property
    def valid(self):
        return [touch for touch in self if touch.valid]


class Touchscreen(object):
    TOUCHSCREEN_EVDEV_NAME = 'raspberrypi-ts'
    EVENT_FORMAT = str('llHHi')
    EVENT_SIZE = struct.calcsize(EVENT_FORMAT)

    def __init__(self, device=None, size=None):
        self._device = device or self.TOUCHSCREEN_EVDEV_NAME
        self.size = size or (800, 480)

        self._running = False
        self._thread = None
        self._f_poll = select.poll()
        self._f_device = io.open(self._touch_device(), 'rb', self.EVENT_SIZE)
        self._f_poll.register(self._f_device, select.POLLIN)
        self.position = Touch(0, 0, 0)
        self.touches = Touches([Touch(x, 0, 0) for x in range(10)])
        self._event_queue = queue.Queue()
        self._touch_slot = 0

        self.surface = pygame.display.set_mode(self.size, pygame.FULLSCREEN)
        pygame.mouse.set_visible(False)

    def _run(self):
        self._running = True
        while self._running:
            self.poll()

    def run(self):
        if self._thread is not None:
            return

        self._thread = threading.Thread(target=self._run)
        self._thread.start()

    def stop(self):
        if self._thread is None:
            return

        self._running = False
        self._thread.join()
        self._thread = None

    @property
    def _current_touch(self):
        return self.touches[self._touch_slot]

    def close(self):
        self._f_device.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.close()

    def __iter__(self):
        pass

    def _lazy_read(self):
        while self._wait_for_events():
            event = self._f_device.read(self.EVENT_SIZE)
            if not event:
                break
            yield event

    def _get_pending_events(self):
        for event in self._lazy_read():
            (tv_sec, tv_usec, type, code, value) = struct.unpack(
                self.EVENT_FORMAT, event)
            self._event_queue.put(
                TouchEvent(tv_sec + (tv_usec / 1000000), type, code, value))

    def _wait_for_events(self, timeout=2):
        return self._f_poll.poll(timeout)

    def poll(self):

        self._get_pending_events()

        while not self._event_queue.empty():
            event = self._event_queue.get()
            self._event_queue.task_done()

            if event.type == EV_SYN:  # Sync
                for touch in self.touches:
                    touch.handle_events()
                return self.touches

            if event.type == EV_ABS:  # Absolute cursor position
                if event.code == ABS_MT_SLOT:
                    self._touch_slot = event.value

                if event.code == ABS_MT_TRACKING_ID:
                    self._current_touch.id = event.value

                if event.code == ABS_MT_POSITION_X:
                    self._current_touch.x = event.value

                if event.code == ABS_MT_POSITION_Y:
                    self._current_touch.y = event.value

                if event.code == ABS_X:
                    self.position.x = event.value

                if event.code == ABS_Y:
                    self.position.y = event.value

        return []

    def _touch_device(self):
        for evdev in glob.glob("/sys/class/input/event*"):
            try:
                with Path(evdev, 'device', 'name').open(mode='r') as f:
                    if f.read().strip() == self._device:
                        return str(Path('/dev', 'input', Path(evdev).name))
            except IOError as e:
                if e.errno != errno.ENOENT:
                    raise
        raise RuntimeError(
            'Unable to locate touchscreen device: {}'.format(self._device))

    def read(self):
        return next(iter(self))


if __name__ == "__main__":
    """ Test for Touchscreen.
    """
    import signal

    ts = Touchscreen()


    def handle_event(event, touch):
        print(["Release", "Press", "Move"][event],
              touch.slot,
              touch._x,
              touch._y)


    for touch in ts.touches:
        touch._on_press = handle_event
        touch._on_release = handle_event
        touch._on_move = handle_event

    ts.run()

    try:
        signal.pause()
    except KeyboardInterrupt:
        print("Stopping thread...")
        ts.stop()
        exit()



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Buttons are widgets that are pressed. They typically update their function
    on release, and the contents can be as simple as a label and as complex
    as a multivalued graph.
"""
import logging
import uuid
from typing import Tuple, Optional

import pygame
from custom_inherit import doc_inherit

from pi_touch_gui._widget_bases import (IButtonWidget, IWidgetInterface)
from pi_touch_gui.widget_functions import ButtonGroupFunction

LOG = logging.getLogger(__name__)


class Button(IButtonWidget):
    """ A simple rectangular button that calls its function on release.

    It is an empty instantiation of IButtonWidget.
    """
    # Button style
    RECTANGLE = 0
    LEFT_ROUND = 1
    RIGHT_ROUND = 2
    OBROUND = 3

    # Button behavior
    CLICK = 0
    HOLD = 1
    TOGGLE = 2

    @doc_inherit(IButtonWidget.__init__, style='numpy_with_merge')
    def __init__(self, position, size,
                 style=None,
                 behavior=None,
                 **kwargs):
        """ Initialize the Button.

        Parameters
        ----------
        style : Int
            The style in which the button is rendered (Button.RECTANGLE is the
            default, but you can also specify LEFT_ROUND, RIGHT_ROUND, or both
            with OBROUND)
        behavior : Int
            The behavior of the button (Button.CLICK is the default, activating
            on release, but you can specify HOLD which activates on press
            and stops on release, or TOGGLE which turns on and off on each
            release).
        """
        super(Button, self).__init__(position, size, **kwargs)
        self.style = style or Button.RECTANGLE
        self.behavior = behavior or Button.CLICK

    @doc_inherit(IButtonWidget.render, style='numpy_with_merge')
    def render(self, page, surface):
        """ Render the button.

        Complicated by the need for one or two rounded ends.  If you play
        your parameters right, you get a circle!
        """
        color, label_color = self.state_colors(self._active)

        radius = int(self._h / 2)
        inset_x = radius if self.style & self.LEFT_ROUND else 0
        inset_w = radius if self.style & self.RIGHT_ROUND else 0
        inset_w = inset_w + inset_x
        subrect = ((self._x + inset_x, self._y),
                   (self._w - inset_w, self._h))
        pygame.draw.rect(surface, color, subrect, 0)
        if self.style & self.LEFT_ROUND:
            pygame.draw.circle(surface, color,
                               (self._x + radius, self._y + radius),
                               radius)
        if self.style & self.RIGHT_ROUND:
            pygame.draw.circle(surface, color,
                               (self._x + self._w - radius, self._y + radius),
                               radius)

        self.render_centered_text(surface, self.label, label_color, color)

    def set(self, page):
        if self.behavior == Button.TOGGLE:
            self._active = True

            if self._function_override:
                return self._function_override.update(page, self, self._active)
            elif self.function:
                return self.function.update(page, self, self._active)

    def clear(self, page):
        if self.behavior == Button.TOGGLE:
            self._active = False

            if self._function_override:
                return self._function_override.update(page, self, self._active)
            elif self.function:
                return self.function.update(page, self, self._active)

    def _on_press(self, touch) -> None:
        """ Hold behavior calls their function on the press event.

        Parameters
        ----------
        touch : Touch
            Touch object; generally not used for buttons
        """
        self.press(touch.page)

    def press(self, page):
        if self.behavior == self.HOLD:
            self._active = True

            if self._function_override:
                return self._function_override.update(page, self, self._active)
            elif self.function:
                self.function.update(page, self, self._active)
        super(Button, self).press(page)

    def _on_release(self, touch):
        return self.release(touch.page)

    def release(self, page):
        if self.behavior == self.TOGGLE:
            self._active = not self._active
        else:
            self._active = False
        return super(Button, self).release(page)


class ButtonGroup(IWidgetInterface):
    """ A group of ToggleButton where zero or one may be active.

    When a button is made active, all the remaining are
    forced to be inactive.

    It is possible for none of the buttons to be active, if you
    de-activate the active button.

    Starts with the first button in the list active.
    """

    @doc_inherit(IWidgetInterface.__init__, style='numpy_with_merge')
    def __init__(self, toggle_buttons, widget_id=None):
        """ Initialize the button group, and hook their button functions.

        Parameters
        ----------
        toggle_buttons : List[ToggleButton]
            An ordered list of ToggleButton objects, where the first button
            is active on startup.
        widget_id : str
            Optional ID for this group, if not given, gets a random uuid.
        """
        self._label = 'group'
        self.widget_id = widget_id or str(uuid.uuid4())

        self.toggle_buttons = toggle_buttons
        self._widgets = list()
        for idx, button in enumerate(toggle_buttons):
            if not isinstance(button, Button):
                raise ValueError(f"Item {idx} {type(button)!r} "
                                 f"is not a Button")
            if button.behavior != Button.TOGGLE:
                raise ValueError(f"Button {idx} is not a TOGGLE")

            # Insert the group control method into the button
            button._function_override = ButtonGroupFunction(self.group_function)
            self._widgets.append(button)

    @property
    def sub_widgets(self):
        return self._widgets

    def group_function(self, page, button, state):
        """ Enforce the group behavior; one button on, all others off.

        Parameters
        ----------
        page : Page
            Page the button is on
        button : ToggleButton
            The button that is receiving the event
        state : bool
            The state being sent to the button

        Returns
        -------
        Optional['Page']
            Returns whatever the button's function returns.  The disabled
            button results are ignored.
        """
        # If we are enabling this button, disable all other buttons
        for btn in self._widgets:
            if btn.id != button.id:
                btn._active = False
                if btn.function:
                    btn.function.update(page, btn, False)

        # Now the actual button press
        btn = next(btn for btn in self._widgets if btn.id == button.id)
        if btn.function:
            return btn.function.update(page, btn, btn._active)
        return None

    @property
    def id(self):
        return self.widget_id

    def event(self, event, touch) -> Tuple[bool, Optional['Page']]:
        for button in self._widgets:
            consumed, new_page = button.event(event, touch)
            if consumed:
                return True, new_page
        return False, None

    def render(self, page, surface):
        for button in self._widgets:
            button.render(page, surface)



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Test - Page with poweroff button
"""

from colors import RED
from page import Page
from pi_touch_gui.widget_buttons import Button
from pi_touch_gui.widget_displays import Label, Light
from pi_touch_gui.widget_functions import ButtonFunction, LightFunction
from tests._utilities import pulsing_light_func, system_func

_exit_label_1 = Label((0, 0), (800, 50), label="Nothing to see here")
_exit_label_2 = Label((0, 50), (800, 430), font_size=100,
                      label_color1=RED, label="... move along!")

_pulse1_light = Light((0, 0), (50, 50),
                      function=LightFunction(pulsing_light_func,
                                             {'rate': 100}))

_pulse2_light = Light((750, 0), (50, 50),
                      function=LightFunction(pulsing_light_func,
                                             {'rate': 105}))

_pulse3_light = Light((0, 430), (50, 50),
                      function=LightFunction(pulsing_light_func,
                                             {'rate': 110}))

_pulse4_light = Light((750, 430), (50, 50),
                      function=LightFunction(pulsing_light_func,
                                             {'rate': 115}))

_off_button = Button((200, 320), (400, 150),
                     style=Button.OBROUND,
                     label="-~= OFF =~-",
                     function=ButtonFunction(system_func,
                                             {'cmd': 'sudo poweroff'}))


def _configure_page(go_to_page):
    """ Custom configuration on the entry page - sets the page and
    button functions.

    Parameters
    ----------
    go_to_page : Callable[[IWidget, bool], Optional['Page']]
    """
    # _poweroff_page.function = go_to_page


def poweroff_page():
    # Tests dynamic Light functions, Label, and default Page function
    _poweroff_page = Page('power_off',
                          [
                              _exit_label_1, _exit_label_2,
                              _pulse1_light, _pulse2_light, _pulse3_light,
                              _pulse4_light,
                              _off_button
                          ],
                          )
    _poweroff_page._config = _configure_page

    return _poweroff_page



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Test - An entry page with pulsing lights
"""

# ---- Pulsing light test function (light)

from colors import RED
from page import Page
from pi_touch_gui.widget_buttons import Button
from pi_touch_gui.widget_displays import Light, Label
from pi_touch_gui.widget_functions import ButtonFunction, LightFunction
from tests._utilities import pulsing_light_func, button_to_page

_left_label = Label((20, 50), (100, 50),
                    label="L", label_alignment=Label.LEFT)
_center_label = Label((20, 100), (100, 50),
                      label="C", label_alignment=Label.CENTER)
_right_label = Label((20, 150), (100, 50),
                     label="R", label_alignment=Label.RIGHT)

_entry_label = Label((0, 0), (800, 50), label="ENTER")

_pulse1_light = Light((0, 0), (50, 50),
                      function=LightFunction(pulsing_light_func,
                                             {'rate': 400}))

_pulse2_light = Light((750, 0), (50, 50),
                      function=LightFunction(pulsing_light_func,
                                             {'rate': 300}))

_pulse3_light = Light((0, 430), (50, 50),
                      function=LightFunction(pulsing_light_func,
                                             {'rate': 200}))

_pulse4_light = Light((750, 430), (50, 50),
                      function=LightFunction(pulsing_light_func,
                                             {'rate': 100}))


def go_to_page_sampler(button, state):
    from tests.test_pages.pages import get_pages
    from tests._utilities import go_to_page
    return go_to_page(get_pages(), 'sampler')


_touch_me_button = Button((200, 150), (400, 100), font_size=100,
                          label_color1=RED, label="Touch Me",
                          function=ButtonFunction(button_to_page,
                                                  {'to_page': 'sampler'}))


def entry_page():
    # Tests dynamic Light functions, Label, and default Page function
    _entry_page = Page('entry', [
        _pulse1_light,
        _entry_label,
        _left_label,
        _center_label,
        _right_label,
        _pulse2_light,
        _pulse3_light,
        _pulse4_light,
        _touch_me_button
    ])
    return _entry_page



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Graphical User Interface objects

    A `Page` is a set of widgets that define the interaction and display
    surface of the GUI

    The `GUI` manages the hardware and runs the Pages
"""
import logging
from typing import NoReturn

import pygame

from pi_touch_gui._utilities import backlog_error
from pi_touch_gui.multitouch.raspberrypi_ts import Touchscreen

# from pi_touch_gui.page import Page

LOG = logging.getLogger(__name__)


class GUI(object):
    # Constants that define the GUI behavior
    FPS = 60
    FADE_TIME = 60
    BLACKOUT_TIME = 300
    BRIGHT_LEVEL = 128
    FADE_LEVEL = 16

    # Initialize the event timer
    last_event = pygame.time.get_ticks()

    def __init__(self, touchscreen, bright=None):
        """ A Graphical User Interface for using a touchscreen.

        Parameters
        ----------
        touchscreen : Touchscreen
            See for example the `multitouch.raspberrypi_ts.Touchscreen`
        bright : int
            Brightness level (to balance visibility and screen wear).  Defaults
            to GUI.BRIGHT_LEVEL.
        """
        self._bright_level = bright or self.BRIGHT_LEVEL
        self._touchscreen = touchscreen
        self._current_page = None
        self._first_page = None
        self._running = False
        self._faded = False
        self._blacked = False

        with open('/sys/class/backlight/rpi_backlight/max_brightness',
                  'r') as f:
            max_str = f.readline()
        self.max_brightness = min(int(max_str), 255)
        self.set_brightness(self._bright_level)

    def reset_display(self):
        """ Enable on the backlight and bring to the default brightness.
        """
        self.last_event = pygame.time.get_ticks()

        if self._faded:
            self.set_brightness(self._bright_level)
            self._faded = False
        if self._blacked:
            self.set_light(True)
            self._blacked = False

    def set_brightness(self, level):
        """ Set the backlight to a determined brightness level.

        Parameters
        ----------
        level : int
        """
        level_str = f"{min(self.max_brightness, level)}\n"
        with open('/sys/class/backlight/rpi_backlight/brightness', 'w') as f:
            f.write(level_str)

    def set_light(self, on):
        """ Turn the backlight activateion to On or Off

        Parameters
        ----------
        on : bool
            True for On, False for Off
        """
        if on is True:
            blank_str = "0"
        else:
            blank_str = "1"
            self._current_page = self._first_page
            self._current_page.render(self._touchscreen.surface)
        with open('/sys/class/backlight/rpi_backlight/bl_power', 'w') as f:
            f.write(blank_str)

    def run(self, start_page, callback=None, cb_rate=None) -> NoReturn:
        """ Run the GUI, staring with the initial Page.

        Does not return, but does an exit() when done. Wrapping
        code in this thread needs to embed this in a Try/Finally to do cleanup.

        Parameters
        ----------
        start_page : Page
            The starting page to display
        callback : Callable[GUI, Page]
            An optional method to inject into the GUI main loop, for managing
            events and activities. This method SHOULD NOT take much time,
            as it is in the middle of the display loop.  Time consuming
            activities should be set up in a separate thread or process, with
            this callback doing quick actions only.
        cb_rate : Int
            The (approximate) number of milliseconds between calls to the
            callback (it helps if it's some multiple of 1/60, as the primary
            loop rate is 60 frames a second).
        """
        self._current_page = start_page
        self._first_page = start_page
        self._running = True

        self._touchscreen.run()

        fps_clock = pygame.time.Clock()

        def handle_touches(e, t):
            """ Call the `Page.touch_handler()` to process an event. """
            if self._blacked:
                # Consume touch if we are fully blacked out
                self.reset_display()
                return
            # Reset the brightness
            self.reset_display()

            new_page = self._current_page.touch_handler(e, t)
            if new_page is not None:
                self._current_page = new_page

        # All touch types go through the `Page.touch_handler()`
        for touch in self._touchscreen.touches:
            touch._on_press = handle_touches
            touch._on_release = handle_touches
            touch._on_move = handle_touches

        try:
            last_hook_ms = pygame.time.get_ticks()
            while self._running:
                # Deal with screen fade and blackout
                time = pygame.time.get_ticks()
                # deltaTime in seconds.
                delta_time = (time - self.last_event) / 1000.0
                if delta_time > self.BLACKOUT_TIME:
                    self.set_light(False)
                    self._blacked = True
                elif delta_time > self.FADE_TIME:
                    self.set_brightness(self.FADE_LEVEL)
                    self._faded = True

                # Call external hook if specified
                if callback is not None and (time - last_hook_ms) > cb_rate:
                    last_hook_ms = time
                    callback(self, self._current_page)

                # Render the current page
                self._current_page.render(self._touchscreen.surface)
                pygame.display.flip()

                # Lock it into the framerate
                fps_clock.tick(self.FPS)

        except Exception as e:
            backlog_error(e, f"Unmanaged Exception {type(e)}: {e}")
        finally:
            self.reset_display()
            LOG.info("Stopping touchscreen thread...")
            self._touchscreen.stop()

            LOG.info("Exiting GUI...")
            exit()

    def stop(self):
        """ Tell the GUI to stop running.

        This method isn't really accessible, but here for completeness.
        """
        self._running = False



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Widget Base Classes, providing the core functionality of the widgets.

    Widgets are single interactive elements attached to a Page.
"""
import math
import uuid
from abc import ABC, abstractmethod
from typing import Tuple, Optional

import pygame
from custom_inherit import doc_inherit
from pygame import Color

from pi_touch_gui._utilities import snapped_value_string
from pi_touch_gui.colors import WHITE, BLACK, MEDIUM_GRAY
from pi_touch_gui.multitouch.raspberrypi_ts import (TS_PRESS, TS_RELEASE,
                                                    TS_MOVE)
from pi_touch_gui.widget_functions import ButtonFunction

# Default font name - specify one here if you want something other
# than the system default.
FONT_NAME = None

# I'm hoping you can init multiple times
pygame.init()


class IWidgetInterface(ABC):
    """ The bare minimum methods to implement to be a widget.
    """

    @property
    @abstractmethod
    def id(self):
        """ The Widget ID, uniquely identifies each widget.
        """

    @property
    def sub_widgets(self):
        return None

    @abstractmethod
    def event(self, event, touch) -> Tuple[bool, Optional['Page']]:
        """ Handle touch events, translating them to individual `on_*` calls.

        Parameters
        ----------
        event : int
            Event code defined for the interface; e.g. _widget_bases.py imports
            TS_PRESS, TS_RELEASE, and TS_MOVE
        touch : Touch
            The Touch representing the event, where `touch.position` is the
            (x, y) coordinate of the touch.

        Returns
        -------
        Optional[Page]
            The page to transition to, or None to stay on this page
        """

    @abstractmethod
    def render(self, page, surface):
        """ Render to the device pygame Surface.

        Parameters
        ----------
        page : Page
        surface : pygame.Surface
        """


class IWidget(IWidgetInterface, ABC):
    """ Widget base class that processes events and holds common state.

        Sub-classes can define any or all of the three event handlers:

            _on_press(Touch) -> None
            _on_release(Touch) -> Optional[Page]
            _on_move(Touch) -> None

        where:
            `Page` : If not None, jump to that page on the next frame.
    """
    # Cache fonts, no need for each widget to have their own copy
    _font_cache = dict()

    def __init__(self, position, size,
                 widget_id=None,
                 font_size=None,
                 label=None,
                 color1=None,
                 color2=None,
                 label_color1=None,
                 label_color2=None):
        """ Initialize the widget.

        Most parameters can be ignored and will pick up sensible defaults.

        Parameters
        ----------
        position : Tuple[int, int]
            The (x, y) top-left corner of the widget. The screen positioning
            is accounted at (0, 0) being the far top-left, increasing down
            and to the right.
        size : Tuple[int, int]
            The (w, h) of the widget's rectangular extent.
            Note that in some widgets labels or indicators could fall outside
            of these extents, and other widgets may choose to re-define the
            extent to something less rectangular.
        widget_id : str
            Optional ID for this widget, if not given, gets a random uuid.
        font_size : int
            Height of the font in pixels.  If not specified, will be 1/2 the
            height of the widget.
        label : str
        color1 : Color
            Main color of the widget as a tuple of (r, g, b) or (r, g, b, a),
            or a pygame.Color; the unselected color or foreground color
        color2 : Color
            Secondary color of the widget; selected color or background color
        label_color1 : Union[Tuple[int, int, int], Tuple[int, int, int, int]]
            Main color of the label; unselected color
        label_color2 : Color
            Secondary color of the label; selected color
        """
        self.widget_id = widget_id or str(uuid.uuid4())

        self.position = position
        self._x, self._y = position

        self.size = size
        self._w, self._h = size

        # Fill in any important values not specified
        self.label = label or ""
        self.color1 = color1 or MEDIUM_GRAY
        self.color2 = color2 or WHITE
        self.label_color1 = label_color1 or WHITE
        self.label_color2 = label_color2 or BLACK

        self.font_size = font_size or int(min(self._w, self._h) >> 1)

        font_key = f"{FONT_NAME}:{self.font_size}"
        if font_key in self._font_cache:
            font = self._font_cache[font_key]
        else:
            font = pygame.font.Font(FONT_NAME, self.font_size)
            self._font_cache[font_key] = font
        self._font = font

        self._pressed = False
        self._touches = []

        # Probe for the existence of the event handlers.
        try:
            callable(self._on_press)
        except AttributeError:
            self._on_press = None
        try:
            callable(self._on_release)
        except AttributeError:
            self._on_release = None
        try:
            callable(self._on_move)
        except AttributeError:
            self._on_move = None
        try:
            callable(self._adjust)
        except AttributeError:
            self._adjust = None

    @property
    def id(self):
        return self.widget_id

    @property
    def pressed(self) -> bool:
        """ True if the widget is currently pressed.
        """
        return len(self._touches) > 0 or self._pressed

    def touch_inside(self, touch) -> bool:
        """ Test a touch against the generic widget rectangle.

        Parameters
        ----------
        touch : Touch
            The touch coordinates, where `touch.position` is a Tuple[int, int]
            of the (x, y) position of the touch.

        Returns
        -------
        bool
            True if the touch is inside the widget extents.
        """
        x, y = touch.position
        return (self._x <= x <= self._x + self._w and
                self._y <= y <= self._y + self._h)

    def event(self, event, touch) -> Tuple[bool, Optional['Page']]:
        # PRESS
        # A press can only be registered when it is in the widget bounds
        if self.touch_inside(touch):
            if event == TS_PRESS and touch not in self._touches:
                self._touches.append(touch)
                if self._on_press:
                    return True, self._on_press(touch)

        # RELEASE
        # A touch can be released even when its not over a widget.
        if event == TS_RELEASE and touch in self._touches:
            self._touches.remove(touch)
            if self._on_release:
                return True, self._on_release(touch)

        # MOVE
        # Touch movement is tracked even when it's not over a widget
        if event == TS_MOVE and touch in self._touches:
            if self._on_move:
                return True, self._on_move(touch)

        # "Not Me" - report not consumed
        return False, None

    def press(self, page):
        self._pressed = True

    def release(self, page):
        self._pressed = False

    def state_colors(self, active=None) -> Tuple[Color, Color]:
        """ Return the (color, label_color) appropriate to the current state.

        Parameters
        ----------
        active : bool
            Override to force the widget to render as active.

        Returns
        -------
        Tuple[Color, Color]
            Returns a tuple of (color, label_color) where the colors can
            be a pygame.Color or (r,g,b) or (r,g,b,a) tuples.
        """
        active = self.pressed or active is True
        if active:
            color = self.color2
            label_color = self.label_color2
        else:
            color = self.color1
            label_color = self.label_color1
        return color, label_color

    def render_centered_text(self, surface, text, color, bg_color=None):
        """ Render text in the center of the widget.

        Parameters
        ----------
        surface : pygame.Surface
        text : str
        color : Color
        bg_color : Color
            If you specify the background color, it can render the text a
            bit faster because it won't use alpha for a clear background.
        """
        if text is None:
            return

        text_bmp = self._font.render(text, 1, color, bg_color)
        textpos = text_bmp.get_rect()
        textpos.centerx = self._x + (self._w / 2)
        textpos.centery = self._y + (self._h / 2)
        surface.blit(text_bmp, textpos)

    def render_left_text(self, surface, text, color, bg_color=None):
        """ Render text in the left side of the widget.

        Parameters
        ----------
        surface : pygame.Surface
        text : str
        color : Color
        bg_color : Color
            If you specify the background color, it can render the text a
            bit faster because it won't use alpha for a clear background.
        """
        if text is None:
            return

        text_bmp = self._font.render(text, 1, color, bg_color)
        textpos = text_bmp.get_rect()
        textpos.left = self._x
        textpos.centery = self._y + (self._h / 2)
        surface.blit(text_bmp, textpos)

    def render_right_text(self, surface, text, color, bg_color=None):
        """ Render text in the right side of the widget.

        Parameters
        ----------
        surface : pygame.Surface
        text : str
        color : Color
        bg_color : Color
            If you specify the background color, it can render the text a
            bit faster because it won't use alpha for a clear background.
        """
        if text is None:
            return

        text_bmp = self._font.render(text, 1, color, bg_color)
        textpos = text_bmp.get_rect()
        textpos.right = self._x + self._w - 1
        textpos.centery = self._y + (self._h / 2)
        surface.blit(text_bmp, textpos)


class IButtonWidget(IWidget):
    """ ButtonWidget base that extends IWidget for push buttons.
    """

    @doc_inherit(IWidget.__init__, style='numpy_with_merge')
    def __init__(self, position, size, function=None, **kwargs):
        """
        Parameters
        ----------
        function : ButtonFunction
            Function to call on button events, with a True/False parameter
            that is True when pressed and False when released.
            Optionally returns a new Page to change pages.
        """
        if (function is not None and
                not issubclass(function.__class__, ButtonFunction)):
            raise ValueError(f"Function type {type(function)} is not a "
                             f"ButtonFunction.")
        self._active = False
        self.function = function
        # Override lets things like the ButtonGroup take over the function,
        # but doesn't serialize so it won't pollute the serial form
        self._function_override = None
        super(IButtonWidget, self).__init__(position, size, **kwargs)

    def _on_release(self, touch) -> Optional['Page']:
        """ Buttons always call their function on the release event.

        It has already been determined that the touch is inside the
        widget via `touch_inside()`.

        Parameters
        ----------
        touch : Touch
            Touch object; generally not used for buttons

        Returns
        -------
        Page
            If specified, the new Page to go to.
        """
        return self.release(touch.page)

    def release(self, page):
        """ Trigger the button, calling its function if specified.
        """
        super(IButtonWidget, self).release(page)
        if self._function_override:
            return self._function_override.update(page, self, self._active)
        elif self.function:
            return self.function.update(page, self, self._active)

    @doc_inherit(IWidget.render, style='numpy_with_merge')
    def render(self, page, surface):
        """ Render the generic button as a filled rectangle with a label.
        """
        color, label_color = self.state_colors()

        pygame.draw.rect(surface, color, (self.position, self.size), 0)
        self.render_centered_text(surface, self.label, label_color, color)


class IControlWidget(IWidget, ABC):
    """ ControlWidget base that extends IWidget for input controls.
    """

    @doc_inherit(IWidget.__init__, style='numpy_with_merge')
    def __init__(self, position, size,
                 min_max,
                 start_value=None,
                 snap_value=None,
                 dynamic=None,
                 **kwargs):
        """
        Parameters
        ----------
        min_max : Tuple[float, float]
            The minimum and maximum values the control might take.
        start_value : float
            The initial value the widget holds.
        snap_value : float
            A value that defines (by example) the display (snap) resolution
            of the widget's value.  E.g. 0.01, 0.1, 1.  The snap value is
            not just a display affect, but controls the resolution of the
            adjusted value of the widget itself.
        dynamic : bool
            True if the function is called each frame the control is active,
            otherwise the function is only called on release.
        """
        super(IControlWidget, self).__init__(position, size,
                                             **kwargs)

        # Constrain the snap values to usable range
        self.snap_value = snap_value or 1
        self.snap_value = 0.01 if self.snap_value < 0.01 else self.snap_value

        self.min_max = min_max
        self._min_val, self._max_val = min_max

        # Value is in 0..1, where adjusted_value is in min_val..max_val
        self._range = self._max_val - self._min_val

        if start_value is not None:
            self._adjusted_value = start_value
            self._value = self.normalize_adjusted_value(start_value)
        else:
            self._adjusted_value = min_max[0]
            self._value = self.normalize_adjusted_value(min_max[0])

        self.dynamic = dynamic or False

        # Function must be set by the child classes AFTER this init
        self.function = None

    def map_value_to_adjusted_value(self, value) -> float:
        """ Map a value in 0..1 to an interpolated value in min_val..max_val.

        Parameters
        ----------
        value : float
            Widget's internal value in 0..1

        Returns
        -------
            Widget's external value in min_val..max_val, snapped to snap_val
        """
        # map [0, 1] to [min_val, max_val] with snap
        interp_value = float(self._min_val + (self._range * value))
        return int(interp_value / self.snap_value + 0.5) * self.snap_value

    def normalize_adjusted_value(self, adjusted_value) -> float:
        """ Normalize a value in min_val..max_val to 0..1

        Parameters
        ----------
        adjusted_value : float
            Widget's external value in min_val..max_val

        Returns
        -------
        float
            Widget's internal value in 0..1
        """
        # map [min_val, max_val] to [0, 1]
        return (adjusted_value - self._min_val) / self._range

    def value_str(self) -> str:
        """ The display version of the widget's external value.

        Returns
        -------
        str
            The string version of self._adjusted_value, which is the value
            the widget represents to the outside world.
        """
        return snapped_value_string(self._adjusted_value, self.snap_value)

    def _on_move(self, touch):
        """ Controls MAY call their function on the move event.

        If the widget was determined to be dynamic, the function is called
        on each move event with a value update.

        Parameters
        ----------
        touch : Touch
            Touch object; generally not used for controls
        """
        if self.dynamic is True and self.function is not None:
            return self.function.update(touch.page, self, self._adjusted_value)

    def _adjust(self, direction):
        """ Adjust the control up or down based on sign(direction)
        """
        step = math.copysign(self.snap_value, direction)
        value = self._adjusted_value + step
        if value < self._min_val:
            value = self._min_val
        elif value > self._max_val:
            value = self._max_val
        self._adjusted_value = value
        self._value = self.normalize_adjusted_value(value)

        if self.function is not None:
            return self.function.update(self, self._adjusted_value)

    def _on_release(self, touch):
        """ Controls call their function on the release event.

        It has already been determined that the touch is inside the
        widget via `touch_inside()`.

        Parameters
        ----------
        touch : Touch
            Touch object; generally not used for controls
        """
        self.release(touch.page)

    def release(self, page):
        """ Trigger the button, calling its function if specified with
        the adjusted value.
        """
        if self.function is not None:
            return self.function.update(page, self, self._adjusted_value)



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Manual Test - set up some pages for viewing and interaction.
"""
import logging

import pygame

from gui import GUI
from page import Page
from pi_touch_gui._utilities import backlog_error
from raspberrypi_ts import Touchscreen
from test_pages.pages import get_pages

logging.basicConfig(
    level=logging.INFO,
    format="[%(asctime)s] [%(levelname)s] "
           "[%(name)s.%(lineno)d] %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S %z"
)
LOG = logging.getLogger(__name__)


def go_to_page(name):
    for page in pages:
        if page.name == name:
            return page
    return None


# Callback ping
def ping(gui, page):
    print(f"PING {page.name!r}")


if __name__ == "__main__":
    pygame.init()

    touchscreen = Touchscreen()
    interface = GUI(touchscreen)

    pages = get_pages()

    for page in pages:
        serialized_page = page.serialize()
        rehydrated_page = Page.deserialize(serialized_page)
        reserialized_page = rehydrated_page.serialize()
        pause = True

    for page in pages:
        functions = page.list_functions()
        print(f"PAGE {page.name!r} calls: {functions}")

    try:
        # Add any hardware or other initialization here
        interface.run(pages[0], callback=ping, cb_rate=1000)
    except Exception as e:
        backlog_error(e, f"Unmanaged exception {type(e)}: {e}")
    finally:
        # Add any hardware or other teardown here
        pass



from pi_touch_gui import Page


def get_pages():
    from test_pages.entry_page import entry_page
    from test_pages.poweroff_page import poweroff_page
    from test_pages.sampler_page import sampler_page

    return [entry_page(), sampler_page(Page.LCARS), poweroff_page()]



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    GUI Colors.

    Convenient named colors
"""

WHITE = 255, 255, 255
LIGHT_GRAY = 192, 192, 192
MEDIUM_GRAY = 128, 128, 128
DARK_GRAY = 64, 64, 64
BLACK = 0, 0, 0

RED = 255, 0, 0
YELLOW = 255, 255, 0
GREEN = 0, 255, 0
CYAN = 0, 255, 255
BLUE = 0, 0, 255
MAGENTA = 255, 0, 255

DARK_RED = 128, 0, 0
DARK_YELLOW = 128, 128, 0
DARK_GREEN = 0, 128, 0
DARK_CYAN = 0, 128, 128
DARK_BLUE = 0, 0, 128
DARK_MAGENTA = 128, 0, 128

# These work well with the LCARS color scheme
LCARS_RED = 206, 102, 102
LCARS_ORANGE = 248, 152, 29
LCARS_PURPLE = 204, 154, 198
LCARS_GRAY_BLUE = 155, 189, 228
LCARS_BEIGE = 255, 204, 153
LCARS_BLUE = 152, 152, 204
LCARS_PEACH = 255, 153, 102
LCARS_PINK = 204, 102, 153



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Testing utility functions that don't fit in any particular class or module.
"""
import math
import os

import pygame

from colors import LCARS_PINK, LCARS_BLUE, WHITE
from test_pages.pages import get_pages


def system_func(**kwargs):
    cmd = kwargs['cmd']
    os.system(cmd)
    return None


def pulsing_light_func(**kwargs):
    """ Provide a dynamic light for testing.

    Specify a `rate` attribute on the Light calling me to change pulse rate.

    Parameters
    ----------
    kwargs : dict

    Returns
    -------
    Tuple(Color, Color)
        The ring, fill colors of the light.
    """
    rate = kwargs['rate']
    ring = WHITE

    time = pygame.time.get_ticks()
    red = math.sin(time / rate) + 1.0  # sets to range 0..2
    fill = (red * 127, 0, 0)

    return ring, fill


def fake_data_graph_function(**kwargs):
    """ Generate two traces of randomized fake data.

    This is a Graph update function.

    Parameters
    ----------
    graph : Graph
    size : int
        The number of samples to return

    Returns
    -------
    List(Dict)
        A list of data dictionaries, where each dict takes the form:
            {"color": <Color>, "data": [<float>]}
    """
    import random

    page = kwargs['page']
    data = page.widget_data
    size = kwargs['num']

    fake_data_1 = data.get('fake_data_1')
    fake_data_2 = data.get('fake_data_2')
    last_update = data.get('last_update', pygame.time.get_ticks())
    fake_num_samples = data.get('fake_num_samples', 1000)

    if fake_data_1 is None:
        fake_data_1 = [5 + random.random()]
        fake_data_2 = [5 + random.random()]
    else:
        time = pygame.time.get_ticks()
        # Update twice per second, to show some life
        if (time - last_update) > 50:
            fake_data_1.append(fake_data_1[-1] + (1.0 * random.random() - 0.5))
            fake_data_2.append(fake_data_2[-1] + (1.0 * random.random() - 0.5))
            last_update = time

    # Truncate to the maximum size to prevent memory crash
    if len(fake_data_1) > fake_num_samples:
        fake_data_1 = fake_data_1[-fake_num_samples:]
        fake_data_2 = fake_data_2[-fake_num_samples:]

    # Make sure the page data is up to date
    data.update({'fake_data_1': fake_data_1,
                 'fake_data_2': fake_data_2,
                 'last_update': last_update,
                 'fake_num_samples': fake_num_samples})

    # Return what was asked for
    return [{"color": LCARS_PINK, "data": fake_data_1[-size:]},
            {"color": LCARS_BLUE, "data": fake_data_2[-size:]}, ]


def go_to_page(pages, name):
    for page in pages:
        if page.name == name:
            return page
    print(f"Page {name!r} not found.")
    return None


def button_to_page(**kwargs):
    page_name = kwargs['to_page']
    return go_to_page(get_pages(), page_name)



# Copyright (c) 2020 Edwin Wise
# MIT License
# See LICENSE for details
"""
    Test - Widget Sample, all the widgets, configured for keyboard grid.
"""

import pygame

from colors import (WHITE, MEDIUM_GRAY, BLUE, RED, BLACK, DARK_RED, GREEN,
                    CYAN, MAGENTA)
from page import Page
from pi_touch_gui.widget_buttons import Button, ButtonGroup
from pi_touch_gui.widget_controls import Dial, Slider
from pi_touch_gui.widget_displays import Indicator, Label, Graph
from pi_touch_gui.widget_functions import (ButtonFunction, IndicatorFunction,
                                           GraphFunction,
                                           DialFunction, SliderFunction,
                                           )
from tests._utilities import fake_data_graph_function, button_to_page


# =========================================================
# Hold-behavior buttons and indicator
# =========================================================

def _up_dn_button_func(**kwargs):
    if kwargs['state']:
        _up_down_flag = kwargs['direction']
    else:
        _up_down_flag = 0

    page = kwargs['page']
    data = page.widget_data
    data['up_down'] = _up_down_flag


def _up_dn_indicator_func(**kwargs):
    page = kwargs['page']
    data = page.widget_data
    _up_down_flag = data.get('up_down', 0)

    if _up_down_flag == 0:
        return False, ''
    elif _up_down_flag == 1:
        return True, "UP"
    else:
        return True, "DN"


_up_button = Button(
    (115, 105), (60, 50),
    label="^",
    behavior=Button.HOLD,
    style=Button.OBROUND,
    function=ButtonFunction(_up_dn_button_func,
                            {'direction': 1}))

_up_dn_indicator = Indicator(
    (115, 160), (60, 50),
    label_color2=WHITE,
    function=IndicatorFunction(_up_dn_indicator_func))

_dn_button = Button(
    (115, 215), (60, 50),
    label="v",
    behavior=Button.HOLD,
    style=Button.OBROUND,
    function=ButtonFunction(_up_dn_button_func,
                            {'direction': -1}))

# =========================================================
# Rounded Button styles
# =========================================================

_back_button = Button((110, 400), (75, 50),
                      label="Back",
                      style=Button.LEFT_ROUND,
                      function=ButtonFunction(button_to_page,
                                              {'to_page': 'entry'}))

_do_nothing_button = Button((190, 400), (150, 50),
                            label="Do Nothing",
                            style=Button.RECTANGLE,
                            color1=MEDIUM_GRAY, color2=WHITE,
                            label_color1=BLUE, label_color2=RED)

_next_button = Button((345, 400), (75, 50),
                      label="Next",
                      style=Button.RIGHT_ROUND,
                      function=ButtonFunction(button_to_page,
                                              {'to_page': 'power_off'}))


# ---- Boop button, for some amusement

def _boop_button_func(**kwargs):
    page = kwargs['page']
    data = page.widget_data
    boop_count = data.get('boop_count', 4) - 1
    data['boop_count'] = boop_count


def _boop_indicator_func(**kwargs):
    """ Countdown; when we reach 0, we exit the program.
    """
    page = kwargs['page']
    data = page.widget_data
    boop_count = data.get('boop_count', 4)
    if boop_count == 4:
        return False, ''
    if boop_count == 0:
        exit(0)
    return True, f"{boop_count}"


_boop_button = Button(
    (430, 400), (50, 50),
    label="X",
    style=Button.OBROUND,
    function=ButtonFunction(_boop_button_func))

_boop_indicator = Indicator(
    (485, 400), (50, 50),
    color1=BLACK, color2=RED,
    label_color1=WHITE, label_color2=BLACK,
    function=IndicatorFunction(_boop_indicator_func))


# =========================================================
# Toggle-behavior Buttons with reset
# =========================================================

def _reset_count_buttons(**kwargs):
    page = kwargs['page']
    _one_button.clear(page)
    _two_button.clear(page)
    _three_button.clear(page)


_count_buttons_label = Label((190, 105), (100, 25),
                             label="Pick Several")
_one_button = Button(
    (190, 130), (100, 50),
    label="One",
    behavior=Button.TOGGLE)
_two_button = Button(
    (190, 185), (100, 50),
    label="Two",
    behavior=Button.TOGGLE)
_three_button = Button(
    (190, 240), (100, 50),
    label="Three",
    behavior=Button.TOGGLE)
_count_reset_button = Button(
    (190, 295), (100, 50),
    label='RESET',
    label_color1=DARK_RED,
    function=ButtonFunction(_reset_count_buttons))

# =========================================================
# Button Group
# =========================================================

_color_buttons_label = Label(
    (300, 105), (100, 25),
    label="Pick One")

_red_button = Button(
    (300, 130), (100, 50),
    label="Red",
    color2=RED,
    behavior=Button.TOGGLE)
_green_button = Button(
    (300, 185), (100, 50),
    label="Green",
    color2=GREEN,
    behavior=Button.TOGGLE)
_blue_button = Button(
    (300, 240), (100, 50), label="Blue",
    color2=BLUE,
    behavior=Button.TOGGLE)

_color_group = ButtonGroup([_red_button, _green_button, _blue_button])


# =========================================================
# Controls and Graphs
# =========================================================

# ---- Graph and Control Functions

def _dial_control_func(**kwargs):
    page = kwargs['page']
    data = page.widget_data
    data['dial_value'] = kwargs['value']


def _slider_h_control_func(**kwargs):
    page = kwargs['page']
    data = page.widget_data
    data['slider_h_value'] = kwargs['value']


def _slider_v_control_func(**kwargs):
    page = kwargs['page']
    data = page.widget_data
    data['slider_v_value'] = kwargs['value']


def _graph_update_func(**kwargs):
    page = kwargs['page']
    data = page.widget_data
    size = kwargs['num']

    last_track_update = data.get('last_track_update', pygame.time.get_ticks())
    track_1_data = data.get('track_1_data', [])
    track_2_data = data.get('track_2_data', [])
    track_3_data = data.get('track_3_data', [])
    max_track_samples = data.get('max_track_samples', 1000)

    dial_value = data.get('dial_value', 0)
    slider_h_value = data.get('slider_h_value', 0)
    slider_v_value = data.get('slider_v_value', 0)

    # Update the graph
    time = pygame.time.get_ticks()
    # Update twice per second, to show some life
    if (time - last_track_update) > 250:
        track_1_data.append(dial_value)
        track_2_data.append(slider_h_value)
        track_3_data.append(slider_v_value)
        last_track_update = time

    # Truncate to the maximum size to prevent memory crash
    if len(track_1_data) > max_track_samples:
        track_1_data = track_1_data[-max_track_samples:]
        track_2_data = track_2_data[-max_track_samples:]
        track_3_data = track_3_data[-max_track_samples:]

    # Make sure the widget data is initialized
    data.update({
        'last_track_update': last_track_update,
        'track_1_data': track_1_data,
        'track_2_data': track_2_data,
        'track_3_data': track_3_data,
        'max_track_samples': max_track_samples
    })

    # Return what was asked for
    return [{"color": CYAN, "data": track_1_data[-size:]},
            {"color": MAGENTA, "data": track_2_data[-size:]},
            {"color": BLUE, "data": track_3_data[-size:]}, ]


# ---- Graph and Control Widgets

_control_graph = Graph(
    (410, 130), (390, 100), min_max=(0, 100),
    function=GraphFunction(_graph_update_func))
_control_dial = Dial((450, 240), (150, 150), min_max=(0, 100),
                     start_value=50, snap_value=1,
                     color2=CYAN,
                     dynamic=True, function=DialFunction(_dial_control_func))
_h_slider = Slider((620, 400), (150, 50),
                   min_max=(25, 75), start_value=40,
                   snap_value=5,
                   label_color1=BLACK, color1=MEDIUM_GRAY, color2=MAGENTA,
                   dynamic=True,
                   function=SliderFunction(_slider_h_control_func))
_v_slider = Slider((620, 240), (50, 150),
                   min_max=(0, 100), start_value=60,
                   snap_value=.1,
                   label_color1=BLACK, color1=MEDIUM_GRAY, color2=BLUE,
                   dynamic=False,
                   function=SliderFunction(_slider_v_control_func))
_fake_graph = Graph((680, 240), (110, 150), min_max=(-100, 100),
                    quiet=True,
                    function=GraphFunction(fake_data_graph_function))


# =========================================================
# Page configuration and definition
# =========================================================


def sampler_page(background):
    # Tests background image, Button, RoundedButton, DynamicButton,
    # ToggleButton, ButtonGroup, Label, Indicator, ...
    _sampler_page = Page('sampler',
                         background=background,
                         widgets=[
                             _up_dn_indicator, _up_button, _dn_button,
                             _back_button, _do_nothing_button, _next_button,
                             _boop_button, _boop_indicator,

                             _count_buttons_label,
                             _one_button, _two_button, _three_button,
                             _count_reset_button,

                             _color_buttons_label,
                             _red_button, _green_button, _blue_button,
                             _color_group,

                             _control_graph, _control_dial, _v_slider,
                             _h_slider,
                             _fake_graph
                         ])

    return _sampler_page
